"""Define test configuration."""
import glob
import re

from pysmartthings.api import (
    API_APP,
    API_APP_OAUTH,
    API_APP_OAUTH_GENERATE,
    API_APP_SETTINGS,
    API_APPS,
    API_BASE,
    API_DEVICE,
    API_DEVICE_COMMAND,
    API_DEVICE_STATUS,
    API_DEVICES,
    API_INSTALLEDAPP,
    API_INSTALLEDAPPS,
    API_LOCATION,
    API_LOCATIONS,
    API_OAUTH_TOKEN,
    API_ROOM,
    API_ROOMS,
    API_SCENE_EXECUTE,
    API_SCENES,
    API_SUBSCRIPTION,
    API_SUBSCRIPTIONS,
    Api,
)
from pysmartthings.smartthings import SmartThings
import pytest

from .utilities import ClientMocker

APP_ID = "c6cde2b0-203e-44cf-a510-3b3ed4706996"
AUTH_TOKEN = "9b3fd445-42d6-441b-b386-99ea51e13cb0"
CLIENT_ID = "7cd4d474-7b36-4e03-bbdb-4cd4ae45a2be"
CLIENT_SECRET = "9b3fd445-42d6-441b-b386-99ea51e13cb0"
DEVICE_ID = "743de49f-036f-4e9c-839a-2f89d57607db"
INSTALLED_APP_ID = "4514eb36-f5fd-4ab2-9520-0597acd1d212"
LOCATION_ID = "397678e5-9995-4a39-9d9f-ae6ba310236b"
ROOM_ID = "7715151d-0314-457a-a82c-5ce48900e065"
REFRESH_TOKEN = "a86a5c8e-0014-44a6-8980-5846633972dd"
SUBSCRIPTION_ID = "7bdf5909-57c4-41f3-9089-e520513bd92a"
SCENE_ID = "9b58411f-5d26-418d-b193-3434a77c484a"

DEVICE_COMMAND_PATTERN = re.compile(r"(device_command_post_[a-z_]+)")


def register_device_commands(mocker):
    """Register all device commands."""
    files = glob.glob("tests/json/device_command_post_*.json")
    for file in files:
        match = DEVICE_COMMAND_PATTERN.search(file)
        if match:
            mocker.post(
                API_DEVICE_COMMAND.format(device_id=DEVICE_ID),
                request=match.group(),
                response={},
            )


def register_url_mocks(mocker):
    """Register the URLs we need to mock."""
    mocker.default_headers = {"Authorization": "Bearer " + AUTH_TOKEN}
    mocker.base_url = API_BASE

    # Locations
    mocker.get(API_LOCATIONS, response="locations")
    mocker.get(API_LOCATION.format(location_id=LOCATION_ID), response="location")

    # Rooms
    mocker.get(API_ROOMS.format(location_id=LOCATION_ID), response="rooms")
    mocker.get(
        API_ROOM.format(location_id=LOCATION_ID, room_id=ROOM_ID), response="room"
    )
    mocker.post(
        API_ROOMS.format(location_id=LOCATION_ID), request="room_post", response="room"
    )
    mocker.put(
        API_ROOM.format(location_id=LOCATION_ID, room_id=ROOM_ID),
        request="room_put",
        response="room",
    )
    mocker.delete(
        API_ROOM.format(location_id=LOCATION_ID, room_id=ROOM_ID), response={}
    )

    # Devices
    mocker.get(API_DEVICES, response="devices")
    mocker.get(
        API_DEVICES,
        response="devices_filtered",
        params=[
            ("locationId", LOCATION_ID),
            ("capability", "switch"),
            ("deviceId", "edd26ac6-d156-4505-9647-3b20118ae4d1"),
            ("deviceId", "be1a61ce-c2a4-4b32-bf8c-31de6d3fa7dd"),
        ],
    )
    mocker.get(API_DEVICE.format(device_id=DEVICE_ID), response="device")
    mocker.get(API_DEVICE_STATUS.format(device_id=DEVICE_ID), response="device_status")

    # Device Commands
    register_device_commands(mocker)

    # Apps
    mocker.get(API_APPS, response="apps")
    mocker.get(API_APP.format(app_id=APP_ID), response="app_get")
    mocker.post(API_APPS, request="app_post_request", response="app_post_response")
    mocker.put(
        API_APP.format(app_id=APP_ID),
        request="app_put_request",
        response="app_put_response",
    )
    mocker.delete(API_APP.format(app_id=APP_ID), response={})
    mocker.get(API_APP_SETTINGS.format(app_id=APP_ID), response="app_settings")
    mocker.put(
        API_APP_SETTINGS.format(app_id=APP_ID),
        request="app_settings",
        response="app_settings",
    )
    mocker.get(API_APP_OAUTH.format(app_id=APP_ID), response="app_oauth_get_response")
    mocker.put(
        API_APP_OAUTH.format(app_id=APP_ID),
        request="app_oauth_put_request",
        response="app_oauth_put_response",
    )
    mocker.post(
        API_APP_OAUTH_GENERATE.format(app_id=APP_ID),
        request="app_oauth_generate_request",
        response="app_oauth_generate_response",
    )

    # InstalledApps
    mocker.get(API_INSTALLEDAPPS, response="installedapps_get_response")
    mocker.request(
        "get",
        "https://api.smartthings.com/installedapps?"
        "currentLocationId=NWMwM2U1MTgtMTE4YS00NGNiLTg1YWQtNzg3N2Qw"
        "YjMwMmU0&currentOffset=MA",
        headers=mocker.default_headers,
        response="installedapps_get_response_2",
    )
    mocker.get(
        API_INSTALLEDAPP.format(installed_app_id=INSTALLED_APP_ID),
        response="installedapp_get_response",
    )
    mocker.delete(
        API_INSTALLEDAPP.format(installed_app_id=INSTALLED_APP_ID),
        response={"count": 1},
    )

    # InstallApp Subscriptions
    mocker.get(
        API_SUBSCRIPTIONS.format(installed_app_id=INSTALLED_APP_ID),
        response="subscriptions_get_response",
    )
    mocker.delete(
        API_SUBSCRIPTIONS.format(installed_app_id=INSTALLED_APP_ID),
        response={"count": 3},
    )
    mocker.post(
        API_SUBSCRIPTIONS.format(installed_app_id=INSTALLED_APP_ID),
        request="subscription_post_request",
        response="subscription_post_response",
    )
    mocker.get(
        API_SUBSCRIPTION.format(
            installed_app_id=INSTALLED_APP_ID, subscription_id=SUBSCRIPTION_ID
        ),
        response="subscription_capability_get_response",
    )
    mocker.get(
        API_SUBSCRIPTION.format(
            installed_app_id=INSTALLED_APP_ID,
            subscription_id="498752fd-db87-4a5e-95f5-25a0e412838d",
        ),
        response="subscription_device_get_response",
    )
    mocker.delete(
        API_SUBSCRIPTION.format(
            installed_app_id=INSTALLED_APP_ID, subscription_id=SUBSCRIPTION_ID
        ),
        response={"count": 1},
    )

    # OAuth Token
    mocker.request("post", API_OAUTH_TOKEN, response="token_response")

    # Scenes
    mocker.get(API_SCENES, response="scenes")
    mocker.get(
        API_SCENES,
        response="scenes_location_filter",
        params=[("locationId", LOCATION_ID)],
    )
    mocker.post(
        API_SCENE_EXECUTE.format(scene_id=SCENE_ID), response={"status": "success"}
    )


@pytest.fixture
def smartthings(event_loop):
    """Fixture for testing against the SmartThings class."""
    # Python 3.5 doesn't support yield in an async method so we have to
    # run the creation and clean-up of the session in the loop manually.
    mocker = ClientMocker()
    register_url_mocks(mocker)
    session = event_loop.run_until_complete(__create_session(event_loop, mocker))
    yield SmartThings(session, AUTH_TOKEN)
    event_loop.run_until_complete(session.close())


@pytest.fixture
def api(event_loop):
    """Fixture for testing against the API."""
    # Python 3.5 doesn't support yield in an async method so we have to
    # run the creation and clean-up of the session in the loop manually.
    mocker = ClientMocker()
    register_url_mocks(mocker)
    session = event_loop.run_until_complete(__create_session(event_loop, mocker))
    yield Api(session, AUTH_TOKEN)
    event_loop.run_until_complete(session.close())


async def __create_session(event_loop, mocker):
    return mocker.create_session(event_loop)



"""Define the subscription module."""

from enum import Enum
from typing import Any, Optional

from .api import Api
from .entity import Entity


class SourceType(Enum):
    """Define the source type of a subscription."""

    UNKNOWN = "UNKNOWN"
    DEVICE = "DEVICE"
    CAPABILITY = "CAPABILITY"


class Subscription:
    """Define the subscription class."""

    def __init__(self):
        """Initialize a new instance of the subscription class."""
        self._subscription_id = None
        self._installed_app_id = None
        self._source_type = SourceType.UNKNOWN
        self._capability = "*"
        self._attribute = "*"
        self._value = "*"
        self._state_change_only = True
        self._subscription_name = None
        # Capability-specific attributes
        self._location_id = None
        # Device-specific attributes
        self._device_id = None
        self._component_id = None

    def apply_data(self, data: dict):
        """Set the states of the app with the supplied data."""
        self._subscription_id = data["id"]
        self._installed_app_id = data["installedAppId"]
        self._source_type = SourceType(data["sourceType"])
        if self._source_type is SourceType.CAPABILITY:
            capability = data["capability"]
            self._location_id = capability["locationId"]
            self._capability = capability["capability"]
            self._attribute = capability.get("attribute", "*")
            self._value = capability.get("value", "*")
            self._state_change_only = capability.get("stateChangeOnly", True)
            self._subscription_name = capability.get("subscriptionName", None)
        if self._source_type is SourceType.DEVICE:
            device = data["device"]
            self._device_id = device["deviceId"]
            self._component_id = device.get("componentId", "*")
            self._capability = device.get("capability", "*")
            self._attribute = device.get("attribute", "*")
            self._value = device.get("value", "*")
            self._state_change_only = device.get("stateChangeOnly", True)
            self._subscription_name = device.get("subscriptionName", None)

    def to_data(self) -> dict:
        """Get a data structure representing this entity."""
        data = {"sourceType": self._source_type.value}
        if self._source_type is SourceType.CAPABILITY:
            capability = {
                "locationId": self._location_id,
                "capability": self._capability,
            }
            if self._attribute and self._attribute != "*":
                capability["attribute"] = self._attribute
            if self._value and self._value != "*":
                capability["value"] = self._value
            if not self._state_change_only:
                capability["stateChangeOnly"] = False
            if self._subscription_name:
                capability["subscriptionName"] = self._subscription_name
            data["capability"] = capability
        if self._source_type is SourceType.DEVICE:
            device = {
                "deviceId": self._device_id,
                "stateChangeOnly": self._state_change_only,
            }
            if self._component_id and self._component_id != "*":
                device["componentId"] = self._component_id
            if self._capability and self._capability != "*":
                device["capability"] = self._capability
            if self._attribute and self._attribute != "*":
                device["attribute"] = self._attribute
            if self._value and self._value != "*":
                device["value"] = self._value
            if self._subscription_name:
                device["subscriptionName"] = self._subscription_name
            data["device"] = device
        return data

    @property
    def subscription_id(self) -> str:
        """Get the id of the subscription."""
        return self._subscription_id

    @property
    def installed_app_id(self) -> str:
        """Get the id of the subscribing app."""
        return self._installed_app_id

    @installed_app_id.setter
    def installed_app_id(self, value: str):
        """Set the id of the subscripting app."""
        self._installed_app_id = value

    @property
    def source_type(self) -> SourceType:
        """Get the type of the event that is being subscribed to."""
        return self._source_type

    @source_type.setter
    def source_type(self, value: Any):
        """Set the typ eof event that is being subscribed to."""
        self._source_type = SourceType(value)

    @property
    def capability(self) -> str:
        """Get the name of the capability that is subscribed."""
        return self._capability

    @capability.setter
    def capability(self, value: str):
        """Get the name of the capability that is subscribed."""
        self._capability = value

    @property
    def attribute(self) -> str:
        """Get the name of the capabilities attribute or * for all."""
        return self._attribute

    @attribute.setter
    def attribute(self, value: str):
        """Set the name of the capabilities attribute or * for all."""
        self._attribute = value

    @property
    def value(self) -> str:
        """Get the value for that will trigger the subscription."""
        return self._value

    @value.setter
    def value(self, value: str):
        """Set the value for that will trigger the subscription."""
        self._value = value

    @property
    def state_change_only(self) -> bool:
        """Get to execute only on a state change."""
        return self._state_change_only

    @state_change_only.setter
    def state_change_only(self, value: bool):
        """Set to execute only on a state change."""
        self._state_change_only = value

    @property
    def subscription_name(self) -> str:
        """Get a name for the subscription."""
        return self._subscription_name

    @subscription_name.setter
    def subscription_name(self, value: str):
        """Set a name for the subscription."""
        self._subscription_name = value

    @property
    def location_id(self) -> str:
        """Get the location id that both the app and source device are in."""
        return self._location_id

    @location_id.setter
    def location_id(self, value: str):
        """Set the location id that both the app and source device are in."""
        self._location_id = value

    @property
    def device_id(self):
        """Get the GUID of the device that is subscribed to."""
        return self._device_id

    @device_id.setter
    def device_id(self, value: str):
        """Set the GUID of the device that is subscribed to."""
        self._device_id = value

    @property
    def component_id(self) -> str:
        """Get the component ID on the device that is subscribed to."""
        return self._component_id

    @component_id.setter
    def component_id(self, value: str):
        """Set the component ID on the device that is subscribed to."""
        self._component_id = value


class SubscriptionEntity(Entity, Subscription):
    """Define a subscription entity."""

    def __init__(self, api: Api, data: Optional[dict] = None):
        """Create a new instance of the SubscriptionEntity class."""
        Entity.__init__(self, api)
        Subscription.__init__(self)
        if data:
            self.apply_data(data)

    async def refresh(self):
        """Refresh the subscription information using the API."""
        data = await self._api.get_subscription(
            self._installed_app_id, self._subscription_id
        )
        self.apply_data(data)

    async def save(self):
        """Subscriptions cannot be updated."""
        raise NotImplementedError



"""Tests for the subscription module."""

from pysmartthings.subscription import SourceType, Subscription, SubscriptionEntity
import pytest

from .conftest import INSTALLED_APP_ID
from .utilities import get_json


class TestSubscription:
    """Tests for the Subscription class."""

    @staticmethod
    def test_init():
        """Test the initialization method."""
        # Arrange/Act
        sub = Subscription()
        # Assert
        assert sub.source_type is SourceType.UNKNOWN
        assert sub.capability == "*"
        assert sub.attribute == "*"
        assert sub.value == "*"
        assert sub.state_change_only

    @staticmethod
    def test_apply_data_capability():
        """Test apply data."""
        # Arrange
        data = get_json("subscription_capability_get_response.json")
        sub = Subscription()
        # Act
        sub.apply_data(data)
        # Assert
        assert sub.subscription_id == "7bdf5909-57c4-41f3-9089-e520513bd92a"
        assert sub.installed_app_id == INSTALLED_APP_ID
        assert sub.source_type == SourceType.CAPABILITY
        assert sub.location_id == "397678e5-9995-4a39-9d9f-ae6ba310236b"
        assert sub.capability == "switchLevel"
        assert sub.attribute == "*"
        assert sub.value == "*"
        assert sub.state_change_only
        assert sub.subscription_name == "switchLevel_sub"

    @staticmethod
    def test_apply_data_device():
        """Test apply data."""
        # Arrange
        data = get_json("subscription_device_get_response.json")
        sub = Subscription()
        # Act
        sub.apply_data(data)
        # Assert
        assert sub.subscription_id == "498752fd-db87-4a5e-95f5-25a0e412838d"
        assert sub.installed_app_id == INSTALLED_APP_ID
        assert sub.source_type == SourceType.DEVICE
        assert sub.device_id == "64e7f664-5b99-4573-b76d-03be3021dc78"
        assert sub.component_id == "*"
        assert sub.capability == "*"
        assert sub.attribute == "*"
        assert sub.value == "*"
        assert sub.state_change_only
        assert not sub.subscription_name

    @staticmethod
    def test_to_data_capability():
        """Test the to_data method for capabilities."""
        # Arrange
        sub = Subscription()
        sub.source_type = SourceType.CAPABILITY
        sub.location_id = "397678e5-9995-4a39-9d9f-ae6ba310236b"
        sub.capability = "switch"
        sub.attribute = "switchLevel"
        sub.value = "100"
        sub.state_change_only = False
        sub.subscription_name = "Test"
        # Act
        data = sub.to_data()
        # Assert
        assert data["sourceType"] == SourceType.CAPABILITY.value
        assert (
            data["capability"]["locationId"] == "397678e5-9995-4a39-9d9f-ae6ba310236b"
        )
        assert data["capability"]["capability"] == "switch"
        assert data["capability"]["attribute"] == "switchLevel"
        assert data["capability"]["value"] == "100"
        assert data["capability"]["subscriptionName"] == "Test"
        assert not data["capability"]["stateChangeOnly"]

    @staticmethod
    def test_to_data_device():
        """Test the to_data method for devices."""
        # Arrange
        sub = Subscription()
        sub.source_type = SourceType.DEVICE
        sub.device_id = "397678e5-9995-4a39-9d9f-ae6ba310236b"
        sub.component_id = "main"
        sub.capability = "switch"
        sub.attribute = "switchLevel"
        sub.value = "100"
        sub.state_change_only = True
        sub.subscription_name = "Test"
        # Act
        data = sub.to_data()
        # Assert
        assert data["sourceType"] == SourceType.DEVICE.value
        assert data["device"]["deviceId"] == "397678e5-9995-4a39-9d9f-ae6ba310236b"
        assert data["device"]["componentId"] == "main"
        assert data["device"]["capability"] == "switch"
        assert data["device"]["attribute"] == "switchLevel"
        assert data["device"]["value"] == "100"
        assert data["device"]["subscriptionName"] == "Test"
        assert data["device"]["stateChangeOnly"]


class TestSubscriptionEntity:
    """Tests for the SubscriptionEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        app = SubscriptionEntity(api)
        app.apply_data(
            {
                "id": "7bdf5909-57c4-41f3-9089-e520513bd92a",
                "installedAppId": INSTALLED_APP_ID,
                "sourceType": "UNKNOWN",
            }
        )
        # Act
        await app.refresh()
        # Assert
        assert app.subscription_name == "switchLevel_sub"

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the refresh method."""
        # Arrange
        app = SubscriptionEntity(api)
        # Act/Assert
        with pytest.raises(NotImplementedError):
            await app.save()



"""Tests for the Device file."""

from pysmartthings.capability import Attribute, Capability
from pysmartthings.device import (
    DEVICE_TYPE_DTH,
    DEVICE_TYPE_UNKNOWN,
    Device,
    DeviceEntity,
    DeviceStatus,
    Status,
)
import pytest

from .conftest import DEVICE_ID, LOCATION_ID, ROOM_ID
from .utilities import get_json


class TestDevice:
    """Tests for the Device class."""

    @staticmethod
    def test_init():
        """Tests whether the Device class initializes correctly."""
        # Arrange/Act
        device = Device()
        # Assert
        assert device.type == DEVICE_TYPE_UNKNOWN
        assert device.capabilities == []
        assert device.components == {}

    @staticmethod
    def test_apply_data():
        """Tests the apply data method."""
        # Arrange
        data = get_json("device.json")
        device = Device()
        # Act
        device.apply_data(data)
        # Assert
        assert device.device_id == DEVICE_ID
        assert device.name == "GE In-Wall Smart Dimmer"
        assert device.label == "Front Porch Lights"
        assert device.location_id == LOCATION_ID
        assert device.room_id == ROOM_ID
        assert device.type == DEVICE_TYPE_DTH
        assert device.device_type_id == "8a9d4b1e3b9b1fe3013b9b206a7f000d"
        assert device.device_type_name == "Dimmer Switch"
        assert device.device_type_network == "ZWAVE"
        assert device.capabilities == [
            "switch",
            "switchLevel",
            "refresh",
            "indicator",
            "sensor",
            "actuator",
            "healthCheck",
            "light",
        ]
        assert device.components == {
            "bottomButton": ["button"],
            "topButton": ["button"],
        }

    @staticmethod
    def test_get_capability():
        """Test the capability retrieval method."""
        # Arrange
        data = get_json("device.json")
        device = Device()
        # Act
        device.apply_data(data)
        # Assert
        assert device.get_capability("switch", "light") == "switch"
        assert device.get_capability("foo", "switch") == "switch"
        assert device.get_capability("foo", "bar") is None


class TestDeviceEntity:
    """Tests for the DeviceEntity class."""

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        await device.refresh()
        # Assert
        assert device.label == "Front Porch Lights"

    @staticmethod
    @pytest.mark.asyncio
    async def test_save(api):
        """Tests the save method."""
        # Arrange
        device = DeviceEntity(api)
        # Act/Assert
        with pytest.raises(NotImplementedError):
            await device.save()

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_on(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.switch_on()
        # Assert
        assert result
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_on_update(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.switch_on(True)
        # Assert
        assert result
        assert device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_off(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.status.switch = True
        # Act
        result = await device.switch_off()
        # Assert
        assert result
        assert device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_switch_off_update(api):
        """Tests the switch_on method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.status.switch = True
        # Act
        result = await device.switch_off(True)
        # Assert
        assert result
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_lock(api):
        """Tests the lock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.lock()
        # Assert
        assert result
        assert not device.status.lock

    @staticmethod
    @pytest.mark.asyncio
    async def test_lock_update(api):
        """Tests the lock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.lock(True)
        # Assert
        assert result
        assert device.status.lock == "locked"

    @staticmethod
    @pytest.mark.asyncio
    async def test_unlock(api):
        """Tests the unlock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.unlock()
        # Assert
        assert result
        assert not device.status.lock

    @staticmethod
    @pytest.mark.asyncio
    async def test_unlock_update(api):
        """Tests the unlock method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.unlock(True)
        # Assert
        assert result
        assert device.status.lock == "unlocked"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_level(api):
        """Tests the set_level method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_level(75, 2)
        # Assert
        assert result
        assert device.status.level == 0
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_level_invalid(api):
        """Tests the set_level method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert level
        levels = [-1, 101]
        for level in levels:
            with pytest.raises(ValueError):
                await device.set_level(level)
        # Assert duration
        with pytest.raises(ValueError):
            await device.set_level(100, -1)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_level_update(api):
        """Tests the set_level method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_level(75, 2, True)
        # Assert
        assert result
        assert device.status.level == 75
        assert device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_speed(api):
        """Tests the set_fan_speed method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_fan_speed(66)
        # Assert
        assert result
        assert device.status.level == 0
        assert not device.status.switch

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_speed_invalid(api):
        """Tests the set_fan_speed method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        with pytest.raises(ValueError):
            await device.set_fan_speed(-1)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_speed_update(api):
        """Tests the set_fan_speed method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_fan_speed(66, True)
        # Assert
        assert result
        assert device.status.fan_speed == 66
        assert device.status.switch

    @staticmethod
    def test_status():
        """Tests the set_level method."""
        # Arrange
        device = DeviceEntity(None, device_id=DEVICE_ID)
        # Act
        status = device.status
        # Assert
        assert status.device_id == DEVICE_ID

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_temperature(api):
        """Tests the set_color_temperature method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color_temperature(3000)
        # Assert
        assert result
        assert device.status.color_temperature == 1

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_temperature_invalid(api):
        """Tests the set_color_temperature method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        values = [0, 30001]
        for value in values:
            with pytest.raises(ValueError):
                await device.set_color_temperature(value)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_temperature_update(api):
        """Tests the set_color_temperature method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color_temperature(3000, True)
        # Assert
        assert result
        assert device.status.color_temperature == 3000

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_hue(api):
        """Tests the set_hue method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_hue(75)
        # Assert
        assert result
        assert device.status.hue == 0

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_hue_invalid(api):
        """Tests the set_hue method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        levels = [-1, 101]
        for level in levels:
            with pytest.raises(ValueError):
                await device.set_hue(level)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_hue_update(api):
        """Tests the set_hue method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_hue(75, True)
        # Assert
        assert result
        assert device.status.hue == 75

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_saturation(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_saturation(75)
        # Assert
        assert result
        assert device.status.saturation == 0

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_saturation_invalid(api):
        """Tests the set_saturation method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        levels = [-1, 101]
        for level in levels:
            with pytest.raises(ValueError):
                await device.set_saturation(level)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_saturation_update(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_saturation(75, True)
        # Assert
        assert result
        assert device.status.saturation == 75

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color(api):
        """Tests the set_color method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(25, 50)
        # Assert
        assert result
        assert device.status.hue == 0
        assert device.status.saturation == 0
        assert device.status.color is None

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_invalid(api):
        """Tests the set_saturation method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                await device.set_color(value, 0)
            with pytest.raises(ValueError):
                await device.set_color(0, value)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_update(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(25, 50, set_status=True)
        # Assert
        assert result
        assert device.status.hue == 25
        assert device.status.saturation == 50
        assert device.status.color == "#4B6432"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_hex(api):
        """Tests the set_color method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(color_hex="#4B6432")
        # Assert
        assert result
        assert device.status.hue == 0
        assert device.status.saturation == 0
        assert device.status.color is None

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_hex_invalid(api):
        """Tests the set_color method invalid values."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Assert
        values = ["000000", "#00000", "#G00000"]
        for value in values:
            with pytest.raises(ValueError):
                await device.set_color(color_hex=value)

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_color_update_hex(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act
        result = await device.set_color(color_hex="#4B6432", set_status=True)
        # Assert
        assert result
        assert device.status.hue == 25
        assert device.status.saturation == 50
        assert device.status.color == "#4B6432"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_fan_mode_legacy(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.thermostat_fan_mode = "on"
        # Act
        result = await device.set_thermostat_fan_mode("auto")
        # Assert
        assert result
        assert device.status.thermostat_fan_mode != "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_fan_mode(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_fan_mode)
        device.status.thermostat_fan_mode = "on"
        # Act
        result = await device.set_thermostat_fan_mode("auto")
        # Assert
        assert result
        assert device.status.thermostat_fan_mode != "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_fan_mode_update(api):
        """Tests the set_saturation method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_fan_mode)
        device.status.thermostat_fan_mode = "on"
        # Act
        result = await device.set_thermostat_fan_mode("auto", set_status=True)
        # Assert
        assert result
        assert device.status.thermostat_fan_mode == "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_mode_legacy(api):
        """Tests the set_thermostat_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.thermostat_mode = "heat"
        # Act
        result = await device.set_thermostat_mode("auto")
        # Assert
        assert result
        assert device.status.thermostat_mode != "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_mode(api):
        """Tests the set_thermostat_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_mode)
        device.status.thermostat_mode = "heat"
        # Act
        result = await device.set_thermostat_mode("auto")
        # Assert
        assert result
        assert device.status.thermostat_mode != "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_thermostat_mode_update(api):
        """Tests the set_thermostat_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_mode)
        device.status.thermostat_mode = "heat"
        # Act
        result = await device.set_thermostat_mode("auto", set_status=True)
        # Assert
        assert result
        assert device.status.thermostat_mode == "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_cooling_setpoint_legacy(api):
        """Tests the set_cooling_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.cooling_setpoint = 70
        # Act
        result = await device.set_cooling_setpoint(76)
        # Assert
        assert result
        assert device.status.cooling_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_cooling_setpoint_mode(api):
        """Tests the set_cooling_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_cooling_setpoint)
        device.status.cooling_setpoint = 70
        # Act
        result = await device.set_cooling_setpoint(76)
        # Assert
        assert result
        assert device.status.cooling_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_cooling_setpoint_update(api):
        """Tests the set_cooling_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_cooling_setpoint)
        device.status.cooling_setpoint = 70
        # Act
        result = await device.set_cooling_setpoint(76, set_status=True)
        # Assert
        assert result
        assert device.status.cooling_setpoint == 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_heating_setpoint_legacy(api):
        """Tests the set_heating_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat)
        device.status.heating_setpoint = 70
        # Act
        result = await device.set_heating_setpoint(76)
        # Assert
        assert result
        assert device.status.heating_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_heating_setpoint_mode(api):
        """Tests the set_heating_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_heating_setpoint)
        device.status.heating_setpoint = 70
        # Act
        result = await device.set_heating_setpoint(76)
        # Assert
        assert result
        assert device.status.heating_setpoint != 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_heating_setpoint_update(api):
        """Tests the set_heating_setpoint method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.thermostat_heating_setpoint)
        device.status.heating_setpoint = 70
        # Act
        result = await device.set_heating_setpoint(76, set_status=True)
        # Assert
        assert result
        assert device.status.heating_setpoint == 76

    @staticmethod
    @pytest.mark.asyncio
    async def test_open(api):
        """Tests the open method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.door_control)
        # Act/Assert
        assert await device.open()
        assert device.status.door is None
        assert await device.open(set_status=True)
        assert device.status.door == "opening"

    @staticmethod
    @pytest.mark.asyncio
    async def test_open_legacy(api):
        """Tests the open method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.garage_door_control)
        # Act/Assert
        assert await device.open()
        assert device.status.door is None
        assert await device.open(set_status=True)
        assert device.status.door == "opening"

    @staticmethod
    @pytest.mark.asyncio
    async def test_close(api):
        """Tests the close method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.door_control)
        # Act/Assert
        assert await device.close()
        assert device.status.door is None
        assert await device.close(set_status=True)
        assert device.status.door == "closing"

    @staticmethod
    @pytest.mark.asyncio
    async def test_close_legacy(api):
        """Tests the close method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.garage_door_control)
        # Act/Assert
        assert await device.close()
        assert device.status.door is None
        assert await device.close(set_status=True)
        assert device.status.door == "closing"

    @staticmethod
    @pytest.mark.asyncio
    async def test_open_window_shade(api):
        """Tests the open method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.window_shade)
        # Act/Assert
        assert await device.open()
        assert device.status.window_shade is None
        assert await device.open(set_status=True)
        assert device.status.window_shade == "opening"

    @staticmethod
    @pytest.mark.asyncio
    async def test_close_window_shade(api):
        """Tests the close method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        device.capabilities.append(Capability.window_shade)
        # Act/Assert
        assert await device.close()
        assert device.status.window_shade is None
        assert await device.close(set_status=True)
        assert device.status.window_shade == "closing"

    @staticmethod
    @pytest.mark.asyncio
    async def test_request_drlc_action(api):
        """Tests the request_drlc_action method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.request_drlc_action(1, 2, "1970-01-01T00:00:00Z", 10, 1)
        assert device.status.drlc_status is None
        assert await device.request_drlc_action(
            1, 2, "1970-01-01T00:00:00Z", 10, 1, set_status=True
        )
        assert device.status.drlc_status == {
            "duration": 10,
            "drlcLevel": 2,
            "start": "1970-01-01T00:00:00Z",
            "override": False,
        }

    @staticmethod
    @pytest.mark.asyncio
    async def test_override_drlc_action(api):
        """Tests the override_drlc_action method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.override_drlc_action(True)
        assert device.status.drlc_status is None
        assert await device.override_drlc_action(True, set_status=True)
        assert device.status.drlc_status == {"override": True}

    @staticmethod
    @pytest.mark.asyncio
    async def test_execute(api):
        """Tests the execute method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.execute("Test", {"data": "Test"})

    @staticmethod
    @pytest.mark.asyncio
    async def test_preset_position(api):
        """Tests the preset_position method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.preset_position()

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_air_conditioner_mode(api):
        """Tests the set_air_conditioner_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.set_air_conditioner_mode("auto")
        assert device.status.air_conditioner_mode is None
        assert await device.set_air_conditioner_mode("auto", set_status=True)
        assert device.status.air_conditioner_mode == "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_fan_mode(api):
        """Tests the set_fan_mode method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.set_fan_mode("auto")
        assert device.status.fan_mode is None
        assert await device.set_fan_mode("auto", set_status=True)
        assert device.status.fan_mode == "auto"

    @staticmethod
    @pytest.mark.asyncio
    async def test_set_air_flow_direction(api):
        """Tests the set_air_flow_direction method."""
        # Arrange
        device = DeviceEntity(api, device_id=DEVICE_ID)
        # Act/Assert
        assert await device.set_air_flow_direction("fixed")
        assert device.status.air_flow_direction is None
        assert await device.set_air_flow_direction("fixed", set_status=True)
        assert device.status.air_flow_direction == "fixed"


class TestDeviceStatus:
    """Tests for the DeviceStatus class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange/Act
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Assert
        assert status.device_id == DEVICE_ID
        assert status.attributes == {}
        assert not status.switch
        assert not status.motion
        assert status.level == 0
        assert status.component_id == "main"

    @staticmethod
    def test_apply_data():
        """Tests the apply_data method."""
        # Arrange
        data = get_json("device_status.json")
        # Act
        status = DeviceStatus(None, DEVICE_ID, data)
        # Assert
        assert len(status.attributes) == 9
        assert status.switch
        assert status.level == 100
        assert len(status.components) == 2
        assert len(status.components["topButton"].attributes) == 3
        assert len(status.components["bottomButton"].attributes) == 3

    @staticmethod
    def test_apply_attribute_update():
        """Tests the apply_attribute_update method."""
        # Arrange
        data = get_json("device_status.json")
        device = DeviceStatus(None, DEVICE_ID, data)
        # Act
        device.apply_attribute_update(
            "main", Capability.switch_level, Attribute.level, 50, "%", {"test": "test"}
        )
        # Assert
        status = device.attributes[Attribute.level]
        assert status.value == 50
        assert status.unit == "%"
        assert status.data == {"test": "test"}

    @staticmethod
    def test_apply_attribute_update_preserve_unit():
        """Tests the apply_attribute_update preserves the old unit."""
        # Arrange
        data = get_json("device_status.json")
        device = DeviceStatus(None, DEVICE_ID, data)
        device.attributes[Capability.switch_level] = Status(40, "%", None)
        # Act
        device.apply_attribute_update(
            "main", Capability.switch_level, Attribute.level, 50
        )
        # Assert
        status = device.attributes[Attribute.level]
        assert status.unit == "%"

    @staticmethod
    def test_apply_attribute_update_child_status():
        """Tests the apply_attribute_update method to a child status."""
        # Arrange
        data = get_json("device_status.json")
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act
        status.apply_attribute_update("bottomButton", "switchLevel", "level", 50)
        # Assert
        assert status.components["bottomButton"].level == 50

    @staticmethod
    def test_values():
        """Test the values property."""
        # Arrange
        data = get_json("device_status.json")
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act/Assert
        assert status.values == {
            "button": None,
            "numberOfButtons": None,
            "supportedButtonValues": None,
            "indicatorStatus": "when off",
            "switch": "on",
            "checkInterval": 1920,
            "healthStatus": None,
            "DeviceWatch-DeviceStatus": None,
            "level": 100,
        }

    @staticmethod
    def test_attributes():
        """Test the attributes property."""
        # Arrange
        data = get_json("device_status.json")
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act/Assert
        assert status.attributes == {
            "button": (None, None, None),
            "numberOfButtons": (None, None, None),
            "supportedButtonValues": (None, None, None),
            "indicatorStatus": ("when off", None, None),
            "switch": ("on", None, None),
            "checkInterval": (
                1920,
                "s",
                {"protocol": "zwave", "hubHardwareId": "000F"},
            ),
            "healthStatus": (None, None, {}),
            "DeviceWatch-DeviceStatus": (None, None, {}),
            "level": (100, "%", None),
        }

    @staticmethod
    def test_attributes_default():
        """Test the attributes property."""
        # Arrange
        data = get_json("device_status.json")
        status = DeviceStatus(None, DEVICE_ID, data)
        # Act/Assert
        assert status.attributes["thermostatSetpoint"] == (None, None, None)

    @staticmethod
    @pytest.mark.asyncio
    async def test_refresh(api):
        """Tests the refresh method."""
        # Arrange
        status = DeviceStatus(api, device_id=DEVICE_ID)
        # Act
        await status.refresh()
        # Assert
        assert len(status.attributes) == 9

    @staticmethod
    def test_switch():
        """Tests the switch property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act
        status.switch = True
        # Assert
        assert status.switch

    @staticmethod
    def test_level():
        """Tests the level property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act
        status.level = 50
        # Assert
        assert status.level == 50

    @staticmethod
    def test_level_range():
        """Tests the level property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                status.level = value

    @staticmethod
    def test_fan_speed():
        """Tests the fan_speed property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act
        status.fan_speed = 50
        # Assert
        assert status.fan_speed == 50

    @staticmethod
    def test_fan_speed_range():
        """Tests the fan_speed property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        with pytest.raises(ValueError):
            status.fan_speed = -1

    @staticmethod
    def test_hue_range():
        """Tests the hue property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                status.hue = value

    @staticmethod
    def test_saturation_range():
        """Tests the hue property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [-1, 101]
        for value in values:
            with pytest.raises(ValueError):
                status.saturation = value

    @staticmethod
    def test_color_temperature_range():
        """Tests the hue property's range."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = [0, 30001]
        for value in values:
            with pytest.raises(ValueError):
                status.color_temperature = value

    @staticmethod
    def test_color_format():
        """Tests the color property's validation."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # Act/Assert
        values = ["000000", "#00000", "#HH2000"]
        for value in values:
            with pytest.raises(ValueError):
                status.color = value

    @staticmethod
    def test_is_on():
        """Tests the is_on method."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)
        status.update_attribute_value(Attribute.acceleration, "active")
        status.level = 100
        # Act/Assert
        assert status.is_on(Attribute.acceleration)
        assert status.is_on(Attribute.level)
        assert not status.is_on(Attribute.switch)

    @staticmethod
    def test_well_known_attributes():
        """Tests the humidity property."""
        # Arrange
        status = DeviceStatus(None, device_id=DEVICE_ID)

        assert status.supported_ac_modes == []
        assert status.supported_ac_fan_modes == []

        status.update_attribute_value(Attribute.humidity, 50)
        status.update_attribute_value(Attribute.temperature, 55)
        status.update_attribute_value(Attribute.thermostat_operating_state, "on")
        status.update_attribute_value(
            Attribute.supported_thermostat_fan_modes, ["on", "off"]
        )
        status.update_attribute_value(
            Attribute.supported_thermostat_modes, ["auto", "off"]
        )
        status.update_attribute_value(Attribute.lock, "locked")
        status.update_attribute_value(Attribute.door, "open")
        status.update_attribute_value(Attribute.window_shade, "closed")
        status.update_attribute_value(Attribute.data, {"test": "test"})
        status.update_attribute_value(Attribute.three_axis, [0, 0, 0])
        status.update_attribute_value(Attribute.supported_ac_modes, ["auto", "cool"])
        status.update_attribute_value(Attribute.fan_mode, "low")
        status.update_attribute_value(Attribute.supported_ac_fan_modes, ["auto", "low"])
        # Act/Assert
        assert status.humidity == 50
        assert status.temperature == 55
        assert status.thermostat_operating_state == "on"
        assert status.supported_thermostat_fan_modes == ["on", "off"]
        assert status.supported_thermostat_modes == ["auto", "off"]
        assert status.lock == "locked"
        assert status.door == "open"
        assert status.window_shade == "closed"
        assert status.data == {"test": "test"}
        assert status.three_axis == [0, 0, 0]
        assert status.supported_ac_modes == ["auto", "cool"]
        assert status.supported_ac_fan_modes == ["auto", "low"]

    @staticmethod
    def test_well_known_ocf_attributes():
        """Tests the OCF related attributes."""
        # Arrange
        data = get_json("device_samsungac_status.json")
        status = DeviceStatus(None, device_id=DEVICE_ID, data=data)
        # Act/Assert
        assert status.ocf_data_model_version == "res.1.1.0,sh.1.1.0"
        assert status.ocf_date_of_manufacture == "2019-02-26T02:05:55Z"
        assert status.ocf_device_id == DEVICE_ID
        assert status.ocf_firmware_version == "0.1.0"
        assert status.ocf_hardware_version == "1.0"
        assert status.ocf_manufacturer_details_link == "http://www.samsung.com"
        assert status.ocf_manufacturer_name == "Samsung Electronics"
        assert (
            status.ocf_model_number
            == "ARTIK051_KRAC_18K|10193441|60010123001111010200000000000000"
        )
        assert status.ocf_name == "Air Conditioner"
        assert status.ocf_os_version == "TizenRT2.0"
        assert status.ocf_platform_id == "d5226d90-1b4f-e59d-5f3f-027ac3b18faf"
        assert status.ocf_platform_version == "0.1.0"
        assert status.ocf_spec_version == "core.1.1.0"
        assert status.ocf_support_link == "http://www.samsung.com/support"
        assert status.ocf_system_time == "02:05:55Z"
        assert status.ocf_vendor_id == "DA-AC-RAC-000001"

    @staticmethod
    def test_well_known_drlc_attributes():
        """Tests the drlc related attributes."""
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # No attribute
        assert status.drlc_status is None
        assert status.drlc_status_duration is None
        assert status.drlc_status_level is None
        assert status.drlc_status_override is None
        assert status.drlc_status_start is None
        # Populated
        drlc_status = {
            "duration": 0,
            "drlcLevel": -1,
            "start": "1970-01-01T00:00:00Z",
            "override": False,
        }
        status.update_attribute_value(Attribute.drlc_status, drlc_status)
        assert status.drlc_status == drlc_status
        assert status.drlc_status_duration == 0
        assert status.drlc_status_level == -1
        assert not status.drlc_status_override
        assert status.drlc_status_start == "1970-01-01T00:00:00Z"
        # Missing
        status.update_attribute_value(Attribute.drlc_status, {})
        assert status.drlc_status == {}
        assert status.drlc_status_duration is None
        assert status.drlc_status_level is None
        assert status.drlc_status_override is None
        assert status.drlc_status_start is None
        # Not valid
        drlc_status = {"duration": "Foo", "drlcLevel": "Foo"}
        status.update_attribute_value(Attribute.drlc_status, drlc_status)
        assert status.drlc_status == drlc_status
        assert status.drlc_status_duration is None
        assert status.drlc_status_level is None

    @staticmethod
    def test_well_known_power_consumption_attributes():
        """Tests the power consumption related attributes."""
        status = DeviceStatus(None, device_id=DEVICE_ID)
        # No attribute
        assert status.power_consumption is None
        assert status.power_consumption_end is None
        assert status.power_consumption_energy is None
        assert status.power_consumption_power is None
        assert status.power_consumption_start is None
        # Populated
        data = {
            "start": "2019-02-24T21:03:04Z",
            "power": 0,
            "energy": 500,
            "end": "2019-02-26T02:05:55Z",
        }
        status.update_attribute_value(Attribute.power_consumption, data)
        assert status.power_consumption == data
        assert status.power_consumption_end == "2019-02-26T02:05:55Z"
        assert status.power_consumption_energy == 500
        assert status.power_consumption_power == 0
        assert status.power_consumption_start == "2019-02-24T21:03:04Z"
        # Missing
        status.update_attribute_value(Attribute.power_consumption, {})
        assert status.power_consumption == {}
        assert status.power_consumption_end is None
        assert status.power_consumption_energy is None
        assert status.power_consumption_power is None
        assert status.power_consumption_start is None
        # Not valid
        data = {"power": "Foo", "energy": "Bar"}
        status.update_attribute_value(Attribute.power_consumption, data)
        assert status.power_consumption == data
        assert status.power_consumption_energy is None
        assert status.power_consumption_power is None



"""
Defines SmartThings capabilities and attributes.

https://smartthings.developer.samsung.com/docs/api-ref/capabilities.html
"""

CAPABILITIES_TO_ATTRIBUTES = {
    "accelerationSensor": ["acceleration"],
    "activityLightingMode": ["lightingMode"],
    "airConditionerFanMode": ["fanMode", "supportedAcFanModes"],
    "airConditionerMode": ["airConditionerMode", "supportedAcModes"],
    "airFlowDirection": ["airFlowDirection"],
    "airQualitySensor": ["airQuality"],
    "alarm": ["alarm"],
    "audioMute": ["mute"],
    "audioVolume": ["volume"],
    "battery": ["battery"],
    "bodyMassIndexMeasurement": ["bmiMeasurement"],
    "bodyWeightMeasurement": ["bodyWeightMeasurement"],
    "button": ["button", "numberOfButtons", "supportedButtonValues"],
    "carbonDioxideMeasurement": ["carbonDioxide"],
    "carbonMonoxideDetector": ["carbonMonoxide"],
    "carbonMonoxideMeasurement": ["carbonMonoxideLevel"],
    "colorControl": ["color", "hue", "saturation"],
    "colorTemperature": ["colorTemperature"],
    "contactSensor": ["contact"],
    "demandResponseLoadControl": ["drlcStatus"],
    "dishwasherMode": ["dishwasherMode"],
    "dishwasherOperatingState": [
        "machineState",
        "supportedMachineStates",
        "dishwasherJobState",
        "completionTime",
    ],
    "doorControl": ["door"],
    "dryerMode": ["dryerMode"],
    "dryerOperatingState": [
        "machineState",
        "supportedMachineStates",
        "dryerJobState",
        "completionTime",
    ],
    "dustSensor": ["fineDustLevel", "dustLevel"],
    "energyMeter": ["energy"],
    "equivalentCarbonDioxideMeasurement": ["equivalentCarbonDioxideMeasurement"],
    "execute": ["data"],
    "fanSpeed": ["fanSpeed"],
    "filterStatus": ["filterStatus"],
    "formaldehydeMeasurement": ["formaldehydeLevel"],
    "garageDoorControl": ["door"],
    "illuminanceMeasurement": ["illuminance"],
    "infraredLevel": ["infraredLevel"],
    "lock": ["lock"],
    "mediaInputSource": ["inputSource", "supportedInputSources"],
    "mediaPlaybackRepeat": ["playbackRepeatMode"],
    "mediaPlaybackShuffle": ["playbackShuffle"],
    "mediaPlayback": ["playbackStatus", "supportedPlaybackCommands"],
    "motionSensor": ["motion"],
    "ocf": [
        "st",
        "mnfv",
        "mndt",
        "mnhw",
        "di",
        "mnsl",
        "dmv",
        "n",
        "vid",
        "mnmo",
        "mnmn",
        "mnml",
        "mnpv",
        "mnos",
        "pi",
        "icv",
    ],
    "odorSensor": ["odorLevel"],
    "ovenMode": ["ovenMode"],
    "ovenOperatingState": [
        "machineState",
        "supportedMachineStates",
        "ovenJobState",
        "completionTime",
        "operationTime",
        "progress",
    ],
    "ovenSetpoint": ["ovenSetpoint"],
    "powerConsumptionReport": ["powerConsumption"],
    "powerMeter": ["power"],
    "powerSource": ["powerSource"],
    "presenceSensor": ["presence"],
    "rapidCooling": ["rapidCooling"],
    "refrigerationSetpoint": ["refrigerationSetpoint"],
    "relativeHumidityMeasurement": ["humidity"],
    "robotCleanerCleaningMode": ["robotCleanerCleaningMode"],
    "robotCleanerMovement": ["robotCleanerMovement"],
    "robotCleanerTurboMode": ["robotCleanerTurboMode"],
    "signalStrength": ["lqi", "rssi"],
    "smokeDetector": ["smoke"],
    "soundSensor": ["sound"],
    "switchLevel": ["level"],
    "switch": ["switch"],
    "tamperAlert": ["tamper"],
    "temperatureMeasurement": ["temperature"],
    "thermostat": [
        "coolingSetpoint",
        "coolingSetpointRange",
        "heatingSetpoint",
        "heatingSetpointRange",
        "schedule",
        "temperature",
        "thermostatFanMode",
        "supportedThermostatFanModes",
        "thermostatMode",
        "supportedThermostatModes",
        "thermostatOperatingState",
        "thermostatSetpoint",
        "thermostatSetpointRange",
    ],
    "thermostatCoolingSetpoint": ["coolingSetpoint"],
    "thermostatFanMode": ["thermostatFanMode", "supportedThermostatFanModes"],
    "thermostatHeatingSetpoint": ["heatingSetpoint"],
    "thermostatMode": ["thermostatMode", "supportedThermostatModes"],
    "thermostatOperatingState": ["thermostatOperatingState"],
    "thermostatSetpoint": ["thermostatSetpoint"],
    "threeAxis": ["threeAxis"],
    "tvChannel": ["tvChannel"],
    "tvocMeasurement": ["tvocLevel"],
    "ultravioletIndex": ["ultravioletIndex"],
    "valve": ["valve"],
    "voltageMeasurement": ["voltage"],
    "washerMode": ["washerMode"],
    "washerOperatingState": [
        "machineState",
        "supportedMachineStates",
        "washerJobState",
        "completionTime",
    ],
    "waterSensor": ["water"],
    "windowShade": ["windowShade"],
}
CAPABILITIES = list(CAPABILITIES_TO_ATTRIBUTES)
ATTRIBUTES = {
    attrib
    for attributes in CAPABILITIES_TO_ATTRIBUTES.values()
    for attrib in attributes
}


class Capability:
    """Define common capabilities."""

    acceleration_sensor = "accelerationSensor"
    activity_lighting_mode = "activityLightingMode"
    air_conditioner_fan_mode = "airConditionerFanMode"
    air_conditioner_mode = "airConditionerMode"
    air_flow_direction = "airFlowDirection"
    air_quality_sensor = "airQualitySensor"
    alarm = "alarm"
    audio_mute = "audioMute"
    audio_volume = "audioVolume"
    battery = "battery"
    body_mass_index_measurement = "bodyMassIndexMeasurement"
    body_weight_measurement = "bodyWeightMeasurement"
    button = "button"
    carbon_dioxide_measurement = "carbonDioxideMeasurement"
    carbon_monoxide_detector = "carbonMonoxideDetector"
    carbon_monoxide_measurement = "carbonMonoxideMeasurement"
    color_control = "colorControl"
    color_temperature = "colorTemperature"
    contact_sensor = "contactSensor"
    demand_response_load_control = "demandResponseLoadControl"
    dishwasher_mode = "dishwasherMode"
    dishwasher_operating_state = "dishwasherOperatingState"
    door_control = "doorControl"
    dryer_mode = "dryerMode"
    dryer_operating_state = "dryerOperatingState"
    dust_sensor = "dustSensor"
    energy_meter = "energyMeter"
    equivalent_carbon_dioxide_measurement = "equivalentCarbonDioxideMeasurement"
    execute = "execute"
    fan_speed = "fanSpeed"
    filter_status = "filterStatus"
    formaldehyde_measurement = "formaldehydeMeasurement"
    garage_door_control = "garageDoorControl"
    illuminance_measurement = "illuminanceMeasurement"
    infrared_level = "infraredLevel"
    lock = "lock"
    media_input_source = "mediaInputSource"
    media_playback = "mediaPlayback"
    media_playback_repeat = "mediaPlaybackRepeat"
    media_playback_shuffle = "mediaPlaybackShuffle"
    motion_sensor = "motionSensor"
    ocf = "ocf"
    odor_sensor = "odorSensor"
    oven_mode = "ovenMode"
    oven_operating_state = "ovenOperatingState"
    oven_setpoint = "ovenSetpoint"
    power_consumption_report = "powerConsumptionReport"
    power_meter = "powerMeter"
    power_source = "powerSource"
    presence_sensor = "presenceSensor"
    rapid_cooling = "rapidCooling"
    refrigeration_setpoint = "refrigerationSetpoint"
    relative_humidity_measurement = "relativeHumidityMeasurement"
    robot_cleaner_cleaning_mode = "robotCleanerCleaningMode"
    robot_cleaner_movement = "robotCleanerMovement"
    robot_cleaner_turbo_mode = "robotCleanerTurboMode"
    signal_strength = "signalStrength"
    smoke_detector = "smokeDetector"
    sound_sensor = "soundSensor"
    switch = "switch"
    switch_level = "switchLevel"
    tamper_alert = "tamperAlert"
    temperature_measurement = "temperatureMeasurement"
    thermostat = "thermostat"
    thermostat_cooling_setpoint = "thermostatCoolingSetpoint"
    thermostat_fan_mode = "thermostatFanMode"
    thermostat_heating_setpoint = "thermostatHeatingSetpoint"
    thermostat_mode = "thermostatMode"
    thermostat_operating_state = "thermostatOperatingState"
    thermostat_setpoint = "thermostatSetpoint"
    three_axis = "threeAxis"
    tv_channel = "tvChannel"
    tvoc_measurement = "tvocMeasurement"
    ultraviolet_index = "ultravioletIndex"
    valve = "valve"
    voltage_measurement = "voltageMeasurement"
    washer_mode = "washerMode"
    washer_operating_state = "washerOperatingState"
    water_sensor = "waterSensor"
    window_shade = "windowShade"


class Attribute:
    """Define common attributes."""

    acceleration = "acceleration"
    air_conditioner_mode = "airConditionerMode"
    air_flow_direction = "airFlowDirection"
    air_quality = "airQuality"
    alarm = "alarm"
    battery = "battery"
    bmi_measurement = "bmiMeasurement"
    body_weight_measurement = "bodyWeightMeasurement"
    button = "button"
    carbon_dioxide = "carbonDioxide"
    carbon_monoxide = "carbonMonoxide"
    carbon_monoxide_level = "carbonMonoxideLevel"
    color = "color"
    color_temperature = "colorTemperature"
    completion_time = "completionTime"
    contact = "contact"
    cooling_setpoint = "coolingSetpoint"
    cooling_setpoint_range = "coolingSetpointRange"
    data = "data"
    di = "di"
    dishwasher_job_state = "dishwasherJobState"
    dishwasher_mode = "dishwasherMode"
    dmv = "dmv"
    door = "door"
    drlc_status = "drlcStatus"
    dryer_job_state = "dryerJobState"
    dryer_mode = "dryerMode"
    dust_level = "dustLevel"
    energy = "energy"
    equivalent_carbon_dioxide_measurement = "equivalentCarbonDioxideMeasurement"
    fan_mode = "fanMode"
    fan_speed = "fanSpeed"
    filter_status = "filterStatus"
    fine_dust_level = "fineDustLevel"
    formaldehyde_level = "formaldehydeLevel"
    heating_setpoint = "heatingSetpoint"
    heating_setpoint_range = "heatingSetpointRange"
    hue = "hue"
    humidity = "humidity"
    icv = "icv"
    illuminance = "illuminance"
    infrared_level = "infraredLevel"
    input_source = "inputSource"
    level = "level"
    lighting_mode = "lightingMode"
    lock = "lock"
    lqi = "lqi"
    machine_state = "machineState"
    mndt = "mndt"
    mnfv = "mnfv"
    mnhw = "mnhw"
    mnml = "mnml"
    mnmn = "mnmn"
    mnmo = "mnmo"
    mnos = "mnos"
    mnpv = "mnpv"
    mnsl = "mnsl"
    motion = "motion"
    mute = "mute"
    n = "n"
    number_of_buttons = "numberOfButtons"
    odor_level = "odorLevel"
    operation_time = "operationTime"
    oven_job_state = "ovenJobState"
    oven_mode = "ovenMode"
    oven_setpoint = "ovenSetpoint"
    pi = "pi"
    playback_repeat_mode = "playbackRepeatMode"
    playback_shuffle = "playbackShuffle"
    playback_status = "playbackStatus"
    power = "power"
    power_consumption = "powerConsumption"
    power_source = "powerSource"
    presence = "presence"
    progress = "progress"
    rapid_cooling = "rapidCooling"
    refrigeration_setpoint = "refrigerationSetpoint"
    robot_cleaner_cleaning_mode = "robotCleanerCleaningMode"
    robot_cleaner_movement = "robotCleanerMovement"
    robot_cleaner_turbo_mode = "robotCleanerTurboMode"
    rssi = "rssi"
    saturation = "saturation"
    schedule = "schedule"
    smoke = "smoke"
    sound = "sound"
    st = "st"
    supported_ac_fan_modes = "supportedAcFanModes"
    supported_ac_modes = "supportedAcModes"
    supported_button_values = "supportedButtonValues"
    supported_input_sources = "supportedInputSources"
    supported_machine_states = "supportedMachineStates"
    supported_playback_commands = "supportedPlaybackCommands"
    supported_thermostat_fan_modes = "supportedThermostatFanModes"
    supported_thermostat_modes = "supportedThermostatModes"
    switch = "switch"
    tamper = "tamper"
    temperature = "temperature"
    thermostat_fan_mode = "thermostatFanMode"
    thermostat_mode = "thermostatMode"
    thermostat_operating_state = "thermostatOperatingState"
    thermostat_setpoint = "thermostatSetpoint"
    thermostat_setpoint_range = "thermostatSetpointRange"
    three_axis = "threeAxis"
    tv_channel = "tvChannel"
    tvoc_level = "tvocLevel"
    ultraviolet_index = "ultravioletIndex"
    valve = "valve"
    vid = "vid"
    voltage = "voltage"
    volume = "volume"
    washer_job_state = "washerJobState"
    washer_mode = "washerMode"
    water = "water"
    window_shade = "windowShade"


ATTRIBUTE_ON_VALUES = {
    Attribute.acceleration: "active",
    Attribute.contact: "open",
    Attribute.filter_status: "replace",
    Attribute.motion: "active",
    Attribute.mute: "muted",
    Attribute.presence: "present",
    Attribute.sound: "detected",
    Attribute.switch: "on",
    Attribute.tamper: "detected",
    Attribute.valve: "open",
    Attribute.water: "wet",
}



"""A python library for interacting with the SmartThings cloud API."""

from .app import (
    APP_TYPE_LAMBDA,
    APP_TYPE_WEBHOOK,
    CLASSIFICATION_AUTOMATION,
    App,
    AppEntity,
    AppOAuth,
    AppOAuthClient,
    AppOAuthClientEntity,
    AppOAuthEntity,
    AppSettings,
    AppSettingsEntity,
)
from .capability import (
    ATTRIBUTES,
    CAPABILITIES,
    CAPABILITIES_TO_ATTRIBUTES,
    Attribute,
    Capability,
)
from .const import __title__, __version__  # noqa
from .device import (
    DEVICE_TYPE_DTH,
    DEVICE_TYPE_ENDPOINT_APP,
    DEVICE_TYPE_OCF,
    DEVICE_TYPE_UNKNOWN,
    DEVICE_TYPE_VIPER,
    Command,
    Device,
    DeviceEntity,
    DeviceStatus,
    DeviceStatusBase,
)
from .errors import APIErrorDetail, APIInvalidGrant, APIResponseError
from .installedapp import (
    InstalledApp,
    InstalledAppEntity,
    InstalledAppStatus,
    InstalledAppType,
)
from .location import Location, LocationEntity
from .oauthtoken import OAuthToken
from .room import Room, RoomEntity
from .scene import Scene, SceneEntity
from .smartthings import SmartThings
from .subscription import SourceType, Subscription, SubscriptionEntity

__all__ = [
    # app
    "APP_TYPE_LAMBDA",
    "APP_TYPE_WEBHOOK",
    "CLASSIFICATION_AUTOMATION",
    "App",
    "AppEntity",
    "AppOAuth",
    "AppOAuthClient",
    "AppOAuthClientEntity",
    "AppOAuthEntity",
    "AppSettings",
    "AppSettingsEntity",
    # capability
    "ATTRIBUTES",
    "CAPABILITIES",
    "CAPABILITIES_TO_ATTRIBUTES",
    "Attribute",
    "Capability",
    # device
    "DEVICE_TYPE_DTH",
    "DEVICE_TYPE_ENDPOINT_APP",
    "DEVICE_TYPE_OCF",
    "DEVICE_TYPE_UNKNOWN",
    "DEVICE_TYPE_VIPER",
    "Command",
    "Device",
    "DeviceEntity",
    "DeviceStatus",
    "DeviceStatusBase",
    # error
    "APIErrorDetail",
    "APIInvalidGrant",
    "APIResponseError",
    # installed app
    "InstalledApp",
    "InstalledAppEntity",
    "InstalledAppStatus",
    "InstalledAppType",
    # location
    "Location",
    "LocationEntity",
    # room
    "Room",
    "RoomEntity",
    # oauthtoken
    "OAuthToken",
    # scene
    "Scene",
    "SceneEntity",
    # smartthings
    "SmartThings",
    # subscription
    "SourceType",
    "Subscription",
    "SubscriptionEntity",
]
